#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers

from gbpservice._i18n import _LI
from gbpservice.contrib.nfp.configurator.lib import constants as const
from gbpservice.contrib.nfp.configurator.lib import demuxer
from gbpservice.contrib.nfp.configurator.lib import utils
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import rpc


LOG = nfp_logging.getLogger(__name__)


class ConfiguratorRpcManager(object):
    """Implements procedure calls invoked by an REST server.

    Implements following RPC methods.
      - create_network_function_device_config
      - delete_network_function_device_config
      - update_network_function_device_config
      - create_network_function_config
      - delete_network_function_config
      - update_network_function_config
      - get_notifications
    Also implements local methods for supporting RPC methods

    """

    def __init__(self, sc, cm, conf, demuxer):
        self.sc = sc
        self.cm = cm
        self.conf = conf
        self.demuxer = demuxer

    def _get_service_agent_instance(self, service_type):
        """Provides service agent instance based on service type.

        :param service_type: firewall/vpn/loadbalancer/generic_config

        Returns: Instance of service agent for a given service type

        """

        return self.cm.sa_instances[service_type]

    def _invoke_service_agent(self, operation,
                              request_data, is_generic_config=False):
        """Maps and invokes an RPC call to a service agent method.

        Takes help of de-multiplexer to get service type and corresponding
        data and invokes the method of service agent. Service agent instance
        is identified based on the service type passed in the request data

        :param operation: Operation type - create/delete/update
        :param request_data: RPC data

        Returns: None

        """

        # Retrieves service type from RPC data
        service_type = self.demuxer.get_service_type(request_data)
        if (const.invalid_service_type == service_type):
            msg = ("Configurator received invalid service type %s." %
                   service_type)
            raise Exception(msg)

        # Retrieves service agent information from RPC data
        # Format of sa_req_list:
        # [{'method': <m1>, 'kwargs': <rpc_data1>}, {}, ... ]
        sa_req_list, service_type = self.demuxer.get_service_agent_info(
            operation, service_type,
            request_data, is_generic_config)
        if not sa_req_list:
            msg = ("Configurator received invalid data format for service"
                   " type %s. Data format: %r" % (service_type, request_data))
            raise Exception(msg)

        # Retrieves service agent instance using service type
        sa_instance = self._get_service_agent_instance(service_type)
        if not sa_instance:
            msg = ("Failed to find agent with service type %s." % service_type)
            raise Exception(msg)

        # Notification data list that needs to be returned after processing
        # RPC request. Format of notification data:
        # notification_data[
        #    {
        #        'receiver': <neutron/orchestrator>,
        #        'resource': <firewall/vpn/loadbalancer/healthmonitor/
        #                    routes/interfaces>,
        #        'method': <network_function_device_notification/
        #                  *aaS response RPC method name>,
        #        'kwargs': [{<data1>}, {data2}]
        #    },
        #    {
        #    }, ...
        # ]
        #
        # Initially, notification data will be empty and is populated
        # after processing each request data in the request data list
        notification_data = {}

        # Handover the request data list and notification data to the
        # identified service agent
        sa_instance.process_request(sa_req_list, notification_data)

    @log_helpers.log_method_call
    def create_network_function_device_config(self, context, request_data):
        """RPC method to configure a network service device.

        Configures a network service VM to facilitate network service
        operation. This RPC method is invoked by the configurator REST
        server. It configures a network service based on the configuration
        request specified in the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC CREATE NETWORK FUNCTION DEVICE CONFIG "
                         "for %(service_type)s, NFI: %(nfi)s, "
                         "NF_ID: %(nf_id)s"),
                     {'service_type': request_data['info']['service_type'],
                      'nfi': request_data['info']['context']['nfi_id'],
                      'nf_id': request_data['info']['context']['nf_id']})

            self._invoke_service_agent('create', request_data, True)
        except Exception as err:
            msg = ("Failed to create network device configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def delete_network_function_device_config(self, context, request_data):
        """RPC method to clear configuration of a network service device.

        Clears configuration of a network service VM. This RPC method is
        invoked by the configurator REST server. It clears configuration
        of a network service based on the configuration request specified
        in the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC DELETE NETWORK FUNCTION DEVICE CONFIG "
                         "for %(service_type)s, NFI: %(nfi)s, "
                         "NF_ID: %(nf_id)s"),
                     {'service_type': request_data['info']['service_type'],
                      'nfi': request_data['info']['context']['nfi_id'],
                      'nf_id': request_data['info']['context']['nf_id']})

            self._invoke_service_agent('delete', request_data, True)
        except Exception as err:
            msg = ("Failed to delete network device configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def update_network_function_device_config(self, context, request_data):
        """RPC method to update of configuration in a network service device.

        Updates configuration of a network service VM. This RPC method is
        invoked by the configurator REST server. It updates configuration
        of a network service based on the configuration request specified
        in the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC UPDATE NETWORK FUNCTION DEVICE CONFIG "
                         "for %(service_type)s, NFI: %(nfi)s, "
                         "NF_ID: %(nf_id)s"),
                     {'service_type': request_data['info']['service_type'],
                      'nfi': request_data['info']['context']['nfi_id'],
                      'nf_id': request_data['info']['context']['nf_id']})

            self._invoke_service_agent('update', request_data, True)
        except Exception as err:
            msg = ("Failed to update network device configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def create_network_function_config(self, context, request_data):
        """RPC method to configure a network service.

        Configures a network service specified in the request data. This
        RPC method is invoked by the configurator REST server. It configures
        a network service based on the configuration request specified in
        the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC CREATE NETWORK FUNCTION CONFIG "
                         "for %(service_type)s "),
                     {'service_type': request_data['info']['service_type']})

            self._invoke_service_agent('create', request_data)
        except Exception as err:
            msg = ("Failed to create network service configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def delete_network_function_config(self, context, request_data):
        """RPC method to clear configuration of a network service.

        Clears configuration of a network service. This RPC method is
        invoked by the configurator REST server. It clears configuration
        of a network service based on the configuration request specified
        in the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC DELETE NETWORK FUNCTION CONFIG "
                         "for %(service_type)s "),
                     {'service_type': request_data['info']['service_type']})

            self._invoke_service_agent('delete', request_data)
        except Exception as err:
            msg = ("Failed to delete network service configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def update_network_function_config(self, context, request_data):
        """RPC method to update of configuration in a network service.

        Updates configuration of a network service. This RPC method is
        invoked by the configurator REST server. It updates configuration
        of a network service based on the configuration request specified
        in the request_data argument.

        :param context: RPC context instance
        :param request_data: RPC data

        Returns: None

        """

        try:
            nfp_context = module_context.init()
            log_info = request_data.get('info')
            logging_context = log_info['context'].get('logging_context', {})
            nfp_context['log_context'] = logging_context
            LOG.info(_LI("Received RPC UPDATE NETWORK FUNCTION CONFIG "
                         "for %(service_type)s "),
                     {'service_type': request_data['info']['service_type']})

            self._invoke_service_agent('update', request_data)
        except Exception as err:
            msg = ("Failed to update network service configuration. %s" %
                   str(err).capitalize())
            LOG.error(msg)

    @log_helpers.log_method_call
    def get_notifications(self, context):
        """RPC method to get all notifications published by configurator.

        Gets all the notifications from the notifications from notification
        queue and sends to configurator agent

        :param context: RPC context instance

        Returns: notification_data

        """
        module_context.init()
        LOG.info(_LI("Received RPC GET NOTIFICATIONS "))
        events = self.sc.get_stashed_events()
        notifications = []
        for event in events:
            notification = event.data
            msg = ("Notification Data: %r" % notification)
            notifications.append(notification)
            LOG.info(msg)
        return notifications


class ConfiguratorModule(object):
    """Implements configurator module APIs.

    Implements methods which are either invoked by registered service
    agents or by the configurator global methods. The methods invoked
    by configurator global methods interface with service agents.

    """

    def __init__(self, sc):
        self.sa_instances = {}
        self.imported_sas = []

    def register_service_agent(self, service_type, service_agent):
        """Stores service agent object.

        :param service_type: Type of service - firewall/vpn/loadbalancer/
        generic_config.
        :param service_agent: Instance of service agent class.

        Returns: Nothing

        """

        if service_type not in self.sa_instances:

            msg = ("Configurator registered service agent of type %s." %
                   service_type)
            LOG.info(msg)
        else:
            msg = ("Identified duplicate registration with service type %s." %
                   service_type)
            LOG.warn(msg)

        # Register the service agent irrespective of previous registration
        self.sa_instances.update({service_type: service_agent})

    def init_service_agents(self, sc, conf):
        """Invokes service agent initialization method.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        Returns: None

        """

        for agent in self.imported_sas:
            try:
                agent.init_agent(self, sc, conf)
            except AttributeError as attr_err:
                LOG.error(agent.__dict__)
                raise AttributeError(agent.__file__ + ': ' + str(attr_err))

    def init_service_agents_complete(self, sc, conf):
        """Invokes service agent initialization complete method.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        Returns: None

        """

        for agent in self.imported_sas:
            try:
                agent.init_agent_complete(self, sc, conf)
            except AttributeError as attr_err:
                LOG.error(agent.__dict__)
                raise AttributeError(agent.__file__ + ': ' + str(attr_err))


def init_rpc(sc, cm, conf, demuxer):
    """Initializes oslo RPC client.

    Creates RPC manager object and registers the configurator's RPC
    agent object with core service controller.

    :param sc: Service Controller object that is used for interfacing
    with core service controller.
    :param cm: Configurator module object that is used for accessing
    ConfiguratorModule class methods.
    :param conf: Configuration object that is used for configuration
    parameter access.
    :param demuxer: De-multiplexer object that is used for accessing
    ServiceAgentDemuxer class methods.

    Returns: None

    """

    # Initializes RPC client
    rpc_mgr = ConfiguratorRpcManager(sc, cm, conf, demuxer)
    configurator_agent = rpc.RpcAgent(sc,
                                      topic=const.CONFIGURATOR_RPC_TOPIC,
                                      manager=rpc_mgr)

    # Registers RPC client object with core service controller
    sc.register_rpc_agents([configurator_agent])


def get_configurator_module_instance(sc, conf):
    """ Provides ConfiguratorModule class object and loads service agents.

    Returns: Instance of ConfiguratorModule class

    """

    cm = ConfiguratorModule(sc)
    conf_utils = utils.ConfiguratorUtils(conf)

    # Loads all the service agents under AGENT_PKG module path
    cm.imported_sas = conf_utils.load_agents(const.AGENTS_PKG)
    msg = ("Configurator loaded service agents from %s location."
           % (cm.imported_sas))
    LOG.info(msg)
    return cm


def nfp_module_init(sc, conf):
    """Initializes configurator module.

    Creates de-multiplexer object and invokes all the agent entry point
    functions. Initializes oslo RPC client for receiving messages from
    REST server. Exceptions are raised to parent function for all types
    of failures.

    :param sc: Service Controller object that is used for interfacing
    with core service controller.
    :param conf: Configuration object that is used for configuration
    parameter access.

    Returns: None
    Raises: Generic exception including error message

    """

    # Create configurator module and de-multiplexer objects
    try:
        cm = get_configurator_module_instance(sc, conf)
        demuxer_instance = demuxer.ServiceAgentDemuxer()
    except Exception as err:
        msg = ("Failed to initialize configurator de-multiplexer. %s."
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Initialized configurator de-multiplexer.")
        LOG.info(msg)

    # Initialize all the pre-loaded service agents
    try:
        cm.init_service_agents(sc, conf)
    except Exception as err:
        msg = ("Failed to initialize configurator agent modules. %s."
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Initialized configurator agents.")
        LOG.info(msg)

    # Initialize RPC client for receiving messages from REST server
    try:
        init_rpc(sc, cm, conf, demuxer_instance)
    except Exception as err:
        msg = ("Failed to initialize configurator RPC with topic %s. %s."
               % (const.CONFIGURATOR_RPC_TOPIC, str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Initialized configurator RPC with topic %s."
               % const.CONFIGURATOR_RPC_TOPIC)
        LOG.debug(msg)


def nfp_module_post_init(sc, conf):
    """Invokes service agent's initialization complete methods.

    :param sc: Service Controller object that is used for interfacing
    with core service controller.
    :param conf: Configuration object that is used for configuration
    parameter access.

    Returns: None
    Raises: Generic exception including error message

    """

    try:
        cm = get_configurator_module_instance(sc, conf)
        cm.init_service_agents_complete(sc, conf)
    except Exception as err:
        msg = ("Failed to trigger initialization complete for configurator"
               " agent modules. %s." % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Initialization of configurator agent modules completed.")
        LOG.info(msg)



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from aim.aim_lib.db import model as aim_lib_model
from aim.api import infra as aim_infra
from aim.api import resource as aim_resource
from aim import context as aim_context
from neutron import context as n_context
from neutron.db import segments_db as segment
from neutron.plugins.ml2 import models as ml2_models
from neutron.tests.unit.extensions import test_securitygroup
from neutron_lib import constants as n_constants
from oslo_config import cfg
import webob.exc

from gbpservice.neutron.db.grouppolicy import group_policy_db as gpdb
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    extension_db as ext_db)
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import db
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (
    aim_validation as av)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_aim_mapping_driver)
from gbpservice.neutron.tests.unit.services.sfc import test_aim_sfc_driver


class AimValidationTestMixin(object):

    def _validate(self):
        # Validate should pass.
        self.assertEqual(api.VALIDATION_PASSED, self.av_mgr.validate())

    def _validate_repair_validate(self):
        # Validate should fail.
        self.assertEqual(
            api.VALIDATION_FAILED_REPAIRABLE, self.av_mgr.validate())

        # Repair.
        self.assertEqual(
            api.VALIDATION_REPAIRED, self.av_mgr.validate(repair=True))

        # Validate should pass.
        self.assertEqual(api.VALIDATION_PASSED, self.av_mgr.validate())

    def _validate_unrepairable(self):
        # Repair should fail.
        self.assertEqual(
            api.VALIDATION_FAILED_UNREPAIRABLE,
            self.av_mgr.validate(repair=True))

    def _validate_fails_binding_ports(self):
        # Repair should fail.
        self.assertEqual(
            api.VALIDATION_FAILED_BINDING_PORTS,
            self.av_mgr.validate(repair=True))

    def _test_aim_resource(self, resource, unexpected_attr_name='name',
                           unexpected_attr_value='unexpected',
                           test_unexpected_monitored=True):
        resource = copy.copy(resource)

        # Make sure the AIM resource exists.
        actual_resource = self.aim_mgr.get(self.aim_ctx, resource)
        self.assertIsNotNone(actual_resource)

        # Only test deleting and modifying if not monitored.
        if not actual_resource.monitored:
            # Delete the AIM resource and test.
            self.aim_mgr.delete(self.aim_ctx, resource)
            self._validate_repair_validate()
            self.assertTrue(
                actual_resource.user_equal(
                    self.aim_mgr.get(self.aim_ctx, resource)))

            # Modify the AIM resource and test.
            self.aim_mgr.update(
                self.aim_ctx, resource, display_name='not what it was')
            self._validate_repair_validate()
            self.assertTrue(
                actual_resource.user_equal(
                    self.aim_mgr.get(self.aim_ctx, resource)))

        # Add unexpected AIM resource and test.
        setattr(resource, unexpected_attr_name, unexpected_attr_value)
        self.aim_mgr.create(self.aim_ctx, resource)
        self._validate_repair_validate()

        if test_unexpected_monitored:
            # Add unexpected monitored AIM resource and test.
            resource.monitored = True
            self.aim_mgr.create(self.aim_ctx, resource)
            self._validate()

            # Delete unexpected monitored AIM resource.
            self.aim_mgr.delete(self.aim_ctx, resource)


class AimValidationTestCase(test_aim_mapping_driver.AIMBaseTestCase,
                            test_securitygroup.SecurityGroupsTestCase,
                            AimValidationTestMixin):

    def setUp(self):
        super(AimValidationTestCase, self).setUp()
        self.av_mgr = av.ValidationManager()
        self.aim_ctx = aim_context.AimContext(self.db_session)


class TestNeutronMapping(AimValidationTestCase):

    def setUp(self):
        super(TestNeutronMapping, self).setUp()

    def _test_routed_subnet(self, subnet_id, gw_ip):
        # Get the AIM Subnet.
        subnet = self._show('subnets', subnet_id)['subnet']
        sn_dn = subnet['apic:distinguished_names'][gw_ip]
        sn = aim_resource.Subnet.from_dn(sn_dn)

        # Test the AIM Subnet.
        self._test_aim_resource(sn, 'gw_ip_mask', '4.3.2.1/24')

    def _test_unscoped_vrf(self, router_id):
        # Get the router's unscoped AIM VRF.
        router = self._show('routers', router_id)['router']
        vrf_dn = router['apic:distinguished_names']['no_scope-VRF']
        vrf = aim_resource.VRF.from_dn(vrf_dn)

        # Test the AIM VRF.
        self._test_aim_resource(vrf)

    def test_static_resources(self):
        # Validate with initial static resources.
        self._validate()

        # Delete the common Tenant and test.
        tenant = aim_resource.Tenant(name='common')
        self.aim_mgr.delete(self.aim_ctx, tenant)
        self._validate_repair_validate()

        # Test unrouted AIM VRF.
        vrf = aim_resource.VRF(
            name=self.driver.aim_mech_driver.apic_system_id + '_UnroutedVRF',
            tenant_name='common')
        self._test_aim_resource(vrf)

        # Test the any Filter.
        filter_name = (self.driver.aim_mech_driver.apic_system_id +
                       '_AnyFilter')
        filter = aim_resource.Filter(
            name=filter_name,
            tenant_name='common')
        self._test_aim_resource(filter)

        # Test the any FilterEntry.
        entry = aim_resource.FilterEntry(
            name='AnyFilterEntry',
            filter_name=filter_name,
            tenant_name='common')
        self._test_aim_resource(entry)

        # Test the default SecurityGroup.
        sg_name = (self.driver.aim_mech_driver.apic_system_id +
                   '_DefaultSecurityGroup')
        sg = aim_resource.SecurityGroup(
            name=sg_name,
            tenant_name='common')
        self._test_aim_resource(sg)

        # Test the default SecurityGroupSubject.
        sg_subject = aim_resource.SecurityGroupSubject(
            name='default',
            security_group_name=sg_name,
            tenant_name='common')
        self._test_aim_resource(sg_subject)

        # Test one default SecurityGroupRule.
        sg_rule = aim_resource.SecurityGroupRule(
            name='arp_egress',
            security_group_subject_name='default',
            security_group_name=sg_name,
            tenant_name='common')
        self._test_aim_resource(sg_rule)

    def _test_project_resources(self, project_id):
        # Validate with initial project resources.
        self._validate()

        # Test AIM Tenant.
        tenant_name = self.driver.aim_mech_driver.name_mapper.project(
            None, project_id)
        tenant = aim_resource.Tenant(name=tenant_name)
        self._test_aim_resource(tenant)

        # Test AIM ApplicationProfile.
        ap = aim_resource.ApplicationProfile(
            tenant_name=tenant_name, name='OpenStack')
        self._test_aim_resource(ap)

    def test_project_resources(self):
        # REVISIT: Currently, a project's AIM Tenant and
        # ApplicationProfile are created in ensure_tenant just before
        # any Neutron/GBP resource is created using that project, and
        # are not cleaned up when the last Neutron/GBP resource
        # needing them is deleted. Instead, they are cleaned up when a
        # notification is received from Keystone that the project has
        # been deleted. We should consider managing these AIM
        # resources more dynamically. If we do, this test will need to
        # be reworked.

        # Test address scope.
        scope = self._make_address_scope(
            self.fmt, 4, name='as1', tenant_id='as_proj')['address_scope']
        self._test_project_resources(scope['project_id'])

        # Test network.
        net_resp = self._make_network(
            self.fmt, 'net1', True, tenant_id='net_proj')
        net = net_resp['network']
        self._test_project_resources(net['project_id'])

        # Test subnet.
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.1.1', '10.0.1.0/24',
            tenant_id='subnet_proj')['subnet']
        self._test_project_resources(subnet['project_id'])

        # Test port. Since Neutron creates the default SG for the
        # port's project even when security_groups=[] is passed, we
        # need to delete the default SG to ensure the port is the only
        # resource owned by port_prog.
        port = self._make_port(
            self.fmt, net['id'], security_groups=[],
            tenant_id='port_proj')['port']
        sgs = self._list(
            'security-groups',
            query_params='project_id=port_proj')['security_groups']
        self.assertEqual(1, len(sgs))
        self._delete('security-groups', sgs[0]['id'])
        self._test_project_resources(port['project_id'])

        # Test security group.
        sg = self._make_security_group(
            self.fmt, 'sg1', 'desc1', tenant_id='sg_proj')['security_group']
        self._test_project_resources(sg['project_id'])

        # Test subnetpool.
        sp = self._make_subnetpool(
            self.fmt, ['10.0.0.0/8'], name='sp1', tenant_id='sp_proj',
            default_prefixlen=24)['subnetpool']
        self._test_project_resources(sp['project_id'])

        # Test router.
        router = self._make_router(
            self.fmt, 'router_proj', 'router1')['router']
        self._test_project_resources(router['project_id'])

        # Test floatingip.
        kwargs = {'router:external': True}
        ext_net_resp = self._make_network(
            self.fmt, 'ext_net', True, arg_list=self.extension_attributes,
            **kwargs)
        ext_net = ext_net_resp['network']
        self._make_subnet(
            self.fmt, ext_net_resp, '100.100.100.1', '100.100.100.0/24')
        fip = self._make_floatingip(
            self.fmt, ext_net['id'], tenant_id='fip_proj')['floatingip']
        self._test_project_resources(fip['project_id'])

    def test_address_scope(self):
        # Create address scope.
        scope4 = self._make_address_scope(
            self.fmt, 4, name='as4')['address_scope']
        scope4_id = scope4['id']
        vrf_dn = scope4['apic:distinguished_names']['VRF']
        self._validate()

        # Delete the address scope's mapping record and test.
        (self.db_session.query(db.AddressScopeMapping).
         filter_by(scope_id=scope4_id).
         delete())
        self._validate_repair_validate()

        # Test AIM VRF.
        vrf = aim_resource.VRF.from_dn(vrf_dn)
        self._test_aim_resource(vrf)

        # Create isomorphic v6 address scope.
        scope6 = self._make_address_scope_for_vrf(
            vrf_dn, n_constants.IP_VERSION_6, name='as6')['address_scope']
        scope6_id = scope6['id']
        self.assertEqual(vrf_dn, scope6['apic:distinguished_names']['VRF'])
        self._validate()

        # Test AIM VRF.
        self._test_aim_resource(vrf)

        # Delete the initial address scope's mapping record and test.
        (self.db_session.query(db.AddressScopeMapping).
         filter_by(scope_id=scope4_id).
         delete())
        self._validate_repair_validate()
        scope4 = self._show('address-scopes', scope4_id)['address_scope']
        self.assertEqual(vrf_dn, scope4['apic:distinguished_names']['VRF'])
        scope6 = self._show('address-scopes', scope6_id)['address_scope']
        self.assertEqual(vrf_dn, scope6['apic:distinguished_names']['VRF'])

        # Test AIM VRF.
        self._test_aim_resource(vrf)

        # Delete the 2nd address scope's mapping record and
        # test. Without this record, there is no way to know that the
        # scopes were previously isomorphic, so they no longer will
        # be isomorphic after repair.
        (self.db_session.query(db.AddressScopeMapping).
         filter_by(scope_id=scope6_id).
         delete())
        self._validate_repair_validate()
        scope4 = self._show('address-scopes', scope4_id)['address_scope']
        self.assertEqual(vrf_dn, scope4['apic:distinguished_names']['VRF'])
        scope6 = self._show('address-scopes', scope6_id)['address_scope']
        scope6_vrf_dn = scope6['apic:distinguished_names']['VRF']
        self.assertNotEqual(vrf_dn, scope6_vrf_dn)

        # Test both AIM VRFs.
        self._test_aim_resource(vrf)
        scope6_vrf = aim_resource.VRF.from_dn(scope6_vrf_dn)
        self._test_aim_resource(scope6_vrf)

    def _test_network_attrs(self, original):
        current = self._show('networks', original['id'])['network']
        self.assertDictEqual(original, current)

    def _test_network_resources(self, net_resp):
        net = net_resp['network']
        net_id = net['id']
        bd_dn = net['apic:distinguished_names']['BridgeDomain']
        epg_dn = net['apic:distinguished_names']['EndpointGroup']

        # Create unrouted subnet.
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.2.1', '10.0.2.0/24')['subnet']
        subnet_id = subnet['id']
        self._validate()
        net = self._show('networks', net['id'])['network']

        # Delete the network's mapping record and test.
        (self.db_session.query(db.NetworkMapping).
         filter_by(network_id=net_id).
         delete())
        self._validate_repair_validate()
        self._test_network_attrs(net)

        # Corrupt the network's mapping record's BD and test.
        with self.db_session.begin():
            mapping = (self.db_session.query(db.NetworkMapping).
                       filter_by(network_id=net_id).
                       one())
            mapping.bd_tenant_name = 'bad_bd_tenant_name'
        self._validate_repair_validate()
        self._test_network_attrs(net)

        # Corrupt the network's mapping record's EPG and test.
        with self.db_session.begin():
            mapping = (self.db_session.query(db.NetworkMapping).
                       filter_by(network_id=net_id).
                       one())
            mapping.epg_app_profile_name = 'bad_epg_app_profilename'
        self._validate_repair_validate()
        self._test_network_attrs(net)

        # Corrupt the network's mapping record's VRF and test.
        with self.db_session.begin():
            mapping = (self.db_session.query(db.NetworkMapping).
                       filter_by(network_id=net_id).
                       one())
            mapping.vrf_name = 'bad_vrf_name'
        self._validate_repair_validate()
        self._test_network_attrs(net)

        # Test AIM BridgeDomain.
        bd = aim_resource.BridgeDomain.from_dn(bd_dn)
        self._test_aim_resource(bd)

        # Test AIM EndpointGroup.
        epg = aim_resource.EndpointGroup.from_dn(epg_dn)
        self._test_aim_resource(epg)

        # Test AIM Subnet.
        if not net['router:external']:
            # Add unexpect AIM Subnet if not external.
            sn = self.driver.aim_mech_driver._map_subnet(
                subnet, '10.0.2.1', bd)
            self.aim_mgr.create(self.aim_ctx, sn)
            self._validate_repair_validate()
        else:
            # Test AIM Subnet if external.
            #
            # REVISIT: If Subnet DN were included in
            # apic:distinguished_names, which it should be, could just
            # use _test_routed_subnet().
            #
            sn = aim_resource.Subnet(
                tenant_name = bd.tenant_name,
                bd_name = bd.name,
                gw_ip_mask='10.0.2.1/24')
            self._test_aim_resource(sn, 'gw_ip_mask', '10.0.3.1/24')

        # Delete subnet extension data and test migration use case.
        (self.db_session.query(ext_db.SubnetExtensionDb).
         filter_by(subnet_id=subnet_id).
         delete())
        self._validate_repair_validate()

        return net

    def test_unrouted_network(self):
        # Create network.
        net_resp = self._make_network(self.fmt, 'net1', True)
        net = net_resp['network']
        net_id = net['id']
        self._validate()

        # Test AIM resources.
        net = self._test_network_resources(net_resp)

        # Delete network extension data and test migration use case.
        (self.db_session.query(ext_db.NetworkExtensionDb).
         filter_by(network_id=net_id).
         delete())
        self._validate_repair_validate()
        self._test_network_attrs(net)

    def _test_external_network(self, vrf_name='openstack_EXT-l1'):
        # Create AIM HostDomainMappingV2.
        hd_mapping = aim_infra.HostDomainMappingV2(
            host_name='*', domain_name='vm2', domain_type='OpenStack')
        self.aim_mgr.create(self.aim_ctx, hd_mapping)

        # Create external network.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l1/instP-n1'}}
        net_resp = self._make_network(
            self.fmt, 'ext_net', True, arg_list=self.extension_attributes,
            **kwargs)
        net = net_resp['network']
        self._validate()

        # Test standard network AIM resources.
        net = self._test_network_resources(net_resp)

        # Test AIM L3Outside.
        l3out = aim_resource.L3Outside(tenant_name='common', name='l1')
        self._test_aim_resource(l3out)
        self._test_network_attrs(net)

        # Test AIM ExternalNetwork.
        en = aim_resource.ExternalNetwork(
            tenant_name='common', l3out_name='l1', name='n1')
        self._test_aim_resource(en)
        self._test_network_attrs(net)

        # Test AIM ExternalSubnet. Note that the NAT strategy code
        # will try to delete unexpected ExternalSubnets even if they
        # are monitored, so we skip that test.
        esn = aim_resource.ExternalSubnet(
            tenant_name='common', l3out_name='l1', external_network_name='n1',
            cidr='0.0.0.0/0')
        self._test_aim_resource(esn, 'cidr', '1.2.3.4/0', False)
        self._test_network_attrs(net)

        # Test AIM VRF.
        vrf = aim_resource.VRF(tenant_name='common', name=vrf_name)
        self._test_aim_resource(vrf)
        self._test_network_attrs(net)

        # Test AIM ApplicationProfile.
        ap = aim_resource.ApplicationProfile(
            tenant_name='common', name='openstack_OpenStack')
        self._test_aim_resource(ap)

        # Test AIM Contract.
        contract = aim_resource.Contract(
            tenant_name='common', name='openstack_EXT-l1')
        self._test_aim_resource(contract)

        # Test AIM ContractSubject.
        subject = aim_resource.ContractSubject(
            tenant_name='common', contract_name='openstack_EXT-l1',
            name='Allow')
        self._test_aim_resource(subject)

        # Test AIM Filter.
        filter = aim_resource.Filter(
            tenant_name='common', name='openstack_EXT-l1')
        self._test_aim_resource(filter)

        # Test AIM FilterEntry.
        entry = aim_resource.FilterEntry(
            tenant_name='common', filter_name='openstack_EXT-l1', name='Any')
        self._test_aim_resource(entry)

        return net

    def test_external_network(self):
        self._test_external_network()

    def test_preexisting_external_network(self):
        # Create pre-existing AIM VRF.
        vrf = aim_resource.VRF(tenant_name='common', name='v1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, vrf)

        # Create pre-existing AIM L3Outside.
        l3out = aim_resource.L3Outside(
            tenant_name='common', name='l1', vrf_name='v1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, l3out)

        # Create pre-existing AIM ExternalNetwork.
        ext_net = aim_resource.ExternalNetwork(
            tenant_name='common', l3out_name='l1', name='n1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, ext_net)

        # Create pre-existing AIM ExternalSubnet.
        ext_sn = aim_resource.ExternalSubnet(
            tenant_name='common', l3out_name='l1', external_network_name='n1',
            cidr='0.0.0.0/0', monitored=True)
        self.aim_mgr.create(self.aim_ctx, ext_sn)

        # Run tests.
        net = self._test_external_network(vrf_name='v1')
        net_id = net['id']

        # Delete network extension data and clear ExternalNetwork
        # contracts to test migration use case.
        (self.db_session.query(ext_db.NetworkExtensionDb).
         filter_by(network_id=net_id).
         delete())
        (self.db_session.query(ext_db.NetworkExtensionCidrDb).
         filter_by(network_id=net_id).
         delete())
        self.aim_mgr.update(
            self.aim_ctx, ext_net,
            provided_contract_names=[],
            consumed_contract_names=[])

        # Test without DN for migration.
        self._validate_unrepairable()

        # Configure invalid DN for migration and test.
        cfg.CONF.set_override(
            'migrate_ext_net_dns', {net_id: 'abcdef'}, group='ml2_apic_aim')
        self._validate_unrepairable()

        # Configure non-existent DN for migration and test.
        cfg.CONF.set_override(
            'migrate_ext_net_dns', {net_id: 'uni/tn-common/out-l9/instP-n1'},
            group='ml2_apic_aim')
        self._validate_unrepairable()

        # Configure correct DN for migration and test.
        cfg.CONF.set_override(
            'migrate_ext_net_dns', {net_id: 'uni/tn-common/out-l1/instP-n1'},
            group='ml2_apic_aim')
        self._validate_repair_validate()
        self._test_network_attrs(net)

    def test_svi_network(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create SVI network.
        kwargs = {'apic:svi': 'True'}
        self._make_network(
            self.fmt, 'net', True, arg_list=self.extension_attributes,
            **kwargs)

        # Test that validation fails.
        self._validate_unrepairable()

    def test_router(self):
        # Create router.
        router = self._make_router(
            self.fmt, self._tenant_id, 'router1')['router']
        contract_dn = router['apic:distinguished_names']['Contract']
        subject_dn = router['apic:distinguished_names']['ContractSubject']
        self._validate()

        # Test AIM Contract.
        contract = aim_resource.Contract.from_dn(contract_dn)
        self._test_aim_resource(contract)

        # Test AIM ContractSubject.
        subject = aim_resource.ContractSubject.from_dn(subject_dn)
        self._test_aim_resource(subject)

    def test_scoped_routing(self):
        # Create shared address scope and subnetpool as tenant_1.
        scope = self._make_address_scope(
            self.fmt, 4, admin=True, name='as1', tenant_id='tenant_1',
            shared=True)['address_scope']
        pool = self._make_subnetpool(
            self.fmt, ['10.0.0.0/8'], admin=True, name='sp1',
            tenant_id='tenant_1', address_scope_id=scope['id'],
            default_prefixlen=24, shared=True)['subnetpool']
        pool_id = pool['id']

        # Create network and subnet as tenant_2.
        net_resp = self._make_network(
            self.fmt, 'net1', True, tenant_id='tenant_2')
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.1.1', '10.0.1.0/24',
            subnetpool_id=pool_id, tenant_id='tenant_2')['subnet']
        subnet_id = subnet['id']

        # Create extra unrouted subnet.
        self._make_subnet(
            self.fmt, net_resp, '10.0.2.1', '10.0.2.0/24',
            subnetpool_id=pool_id, tenant_id='tenant_2')

        # Create external network.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l1/instP-n1'}}
        ext_net = self._make_network(
            self.fmt, 'ext_net', True, arg_list=self.extension_attributes,
            **kwargs)['network']

        # Create extra external network to test CloneL3Out record below.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l2/instP-n2'}}
        self._make_network(
            self.fmt, 'extra_ext_net', True,
            arg_list=self.extension_attributes, **kwargs)

        # Create router as tenant_2.
        kwargs = {'apic:external_provided_contracts': ['p1', 'p2'],
                  'apic:external_consumed_contracts': ['c1', 'c2'],
                  'external_gateway_info': {'network_id': ext_net['id']}}
        router = self._make_router(
            self.fmt, 'tenant_2', 'router1',
            arg_list=self.extension_attributes, **kwargs)['router']
        router_id = router['id']

        # Validate before adding subnet to router.
        self._validate()

        # Add subnet to router.
        self.l3_plugin.add_router_interface(
            n_context.get_admin_context(), router_id,
            {'subnet_id': subnet_id})
        self._validate()

        # Test AIM Subnet.
        self._test_routed_subnet(subnet_id, '10.0.1.1')

        # Determine clone L3Outside identity based on VRF.
        vrf_dn = scope['apic:distinguished_names']['VRF']
        vrf = aim_resource.VRF.from_dn(vrf_dn)
        tenant_name = vrf.tenant_name
        l3out_name = 'l1-%s' % vrf.name

        # Test AIM L3Outside.
        l3out = aim_resource.L3Outside(
            tenant_name=tenant_name, name=l3out_name)
        self._test_aim_resource(l3out)

        # Test AIM ExternalNetwork.
        en = aim_resource.ExternalNetwork(
            tenant_name=tenant_name, l3out_name=l3out_name, name='n1')
        self._test_aim_resource(en)

        # Test AIM ExternalSubnet.
        esn = aim_resource.ExternalSubnet(
            tenant_name=tenant_name, l3out_name=l3out_name,
            external_network_name='n1', cidr='0.0.0.0/0')
        self._test_aim_resource(esn, 'cidr', '1.2.3.4/0')

        # Delete the CloneL3Out record and test.
        (self.db_session.query(aim_lib_model.CloneL3Out).
         filter_by(tenant_name=tenant_name, name=l3out_name).
         delete())
        self._validate_repair_validate()

        # Corrupt the CloneL3Out record and test.
        with self.db_session.begin():
            record = (self.db_session.query(aim_lib_model.CloneL3Out).
                      filter_by(tenant_name=tenant_name, name=l3out_name).
                      one())
            record.source_name = 'l2'
        self._validate_repair_validate()

        # Add monitored L3Out and unexpected CloneL3Out record and test.
        with self.db_session.begin():
            unexpected_l3out_name = 'l2-%s' % vrf.name
            unexpected_l3out = aim_resource.L3Outside(
                tenant_name=tenant_name, name=unexpected_l3out_name,
                monitored=True)
            self.aim_mgr.create(self.aim_ctx, unexpected_l3out)
            record = aim_lib_model.CloneL3Out(
                source_tenant_name='common', source_name='l2',
                name=unexpected_l3out_name, tenant_name=tenant_name)
            self.db_session.add(record)
        self._validate_repair_validate()

    def test_unscoped_routing(self):
        # Create shared network and unscoped subnet as tenant_1.
        net_resp = self._make_network(
            self.fmt, 'net1', True, tenant_id='tenant_1', shared=True)
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.1.1', '10.0.1.0/24',
            tenant_id='tenant_1')['subnet']
        subnet1_id = subnet['id']

        # Create unshared network and unscoped subnet as tenant_2.
        net_resp = self._make_network(
            self.fmt, 'net2', True, tenant_id='tenant_2')
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.2.1', '10.0.2.0/24',
            tenant_id='tenant_2')['subnet']
        subnet2_id = subnet['id']

        # Create extra unrouted subnet.
        self._make_subnet(
            self.fmt, net_resp, '10.0.3.1', '10.0.3.0/24',
            tenant_id='tenant_2')

        # Create external network.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l1/instP-n1'}}
        ext_net = self._make_network(
            self.fmt, 'ext_net', True, arg_list=self.extension_attributes,
            **kwargs)['network']

        # Create router as tenant_2.
        kwargs = {'apic:external_provided_contracts': ['p1', 'p2'],
                  'apic:external_consumed_contracts': ['c1', 'c2'],
                  'external_gateway_info': {'network_id': ext_net['id']}}
        router = self._make_router(
            self.fmt, 'tenant_2', 'router1',
            arg_list=self.extension_attributes, **kwargs)['router']
        router_id = router['id']

        # Validate before adding subnet to router.
        self._validate()

        # Add unshared subnet to router.
        self.l3_plugin.add_router_interface(
            n_context.get_admin_context(), router_id,
            {'subnet_id': subnet2_id})
        self._validate()

        # Test AIM Subnet and VRF.
        self._test_routed_subnet(subnet2_id, '10.0.2.1')
        self._test_unscoped_vrf(router_id)

        # Add shared subnet to router.
        self.l3_plugin.add_router_interface(
            n_context.get_admin_context(), router_id,
            {'subnet_id': subnet1_id})
        self._validate()

        # Test AIM Subnets and VRF.
        self._test_routed_subnet(subnet2_id, '10.0.2.1')
        self._test_routed_subnet(subnet1_id, '10.0.1.1')
        self._test_unscoped_vrf(router_id)

    def test_security_group(self):
        # Create security group with a rule.
        sg = self._make_security_group(
            self.fmt, 'sg1', 'security group 1')['security_group']
        rule1 = self._build_security_group_rule(
            sg['id'], 'ingress', 'tcp', '22', '23')
        rules = {'security_group_rules': [rule1['security_group_rule']]}
        sg_rule = self._make_security_group_rule(
            self.fmt, rules)['security_group_rules'][0]
        self._validate()

        # Test the AIM SecurityGroup.
        tenant_name = self.driver.aim_mech_driver.name_mapper.project(
            None, sg['project_id'])
        sg_name = sg['id']
        aim_sg = aim_resource.SecurityGroup(
            name=sg_name, tenant_name=tenant_name)
        self._test_aim_resource(aim_sg)

        # Test the AIM SecurityGroupSubject.
        aim_subject = aim_resource.SecurityGroupSubject(
            name='default', security_group_name=sg_name,
            tenant_name=tenant_name)
        self._test_aim_resource(aim_subject)

        # Test the AIM SecurityGroupRule.
        aim_rule = aim_resource.SecurityGroupRule(
            name=sg_rule['id'],
            security_group_subject_name='default',
            security_group_name=sg_name,
            tenant_name=tenant_name)
        self._test_aim_resource(aim_rule)

    def test_network_segment(self):
        # REVISIT: Test repair when migration from other types to
        # 'opflex' is supported.

        # Create network.
        net = self._make_network(self.fmt, 'net1', True)['network']

        # Change network's segment to an unknown type.
        self.db_session.query(segment.NetworkSegment).filter_by(
            network_id=net['id']).update({'network_type': 'xxx'})

        # Test that validation fails.
        self._validate_unrepairable()

    def test_port_binding(self):
        # Create network, subnet, and bound port.
        net_resp = self._make_network(self.fmt, 'net1', True)
        net = net_resp['network']
        subnet = self._make_subnet(
            self.fmt, net_resp, None, '10.0.0.0/24')['subnet']
        fixed_ips = [{'subnet_id': subnet['id'], 'ip_address': '10.0.0.100'}]
        port = self._make_port(
            self.fmt, net['id'], fixed_ips=fixed_ips)['port']
        port = self._bind_port_to_host(port['id'], 'host1')['port']
        self._validate()

        # Change port binding level to unknown mechanism driver and
        # test.
        self.db_session.query(ml2_models.PortBindingLevel).filter_by(
            port_id=port['id'], level=0).update({'driver': 'xxx'})
        self._validate_repair_validate()

        # Change port binding level to incorrect host and test.
        self.db_session.query(ml2_models.PortBindingLevel).filter_by(
            port_id=port['id'], level=0).update({'host': 'yyy'})
        self._validate_repair_validate()

        # Change port binding level to null segment ID and test.
        self.db_session.query(ml2_models.PortBindingLevel).filter_by(
            port_id=port['id'], level=0).update({'segment_id': None})
        self._validate_repair_validate()

        # Change port binding level to unknown mechanism driver, set
        # bad host, and test that repair fails.
        #
        # REVISIT: The apic_aim MD currently tries to allocate a
        # dynamic segment whenever there is no agent on the port's
        # host, which is probably wrong, but it does fail to bind, so
        # this test succeeds.
        self.db_session.query(ml2_models.PortBindingLevel).filter_by(
            port_id=port['id'], level=0).update({'driver': 'xxx',
                                                 'host': 'yyy'})
        self.db_session.query(ml2_models.PortBinding).filter_by(
            port_id=port['id']).update({'host': 'yyy'})
        self._validate_fails_binding_ports()

    def test_legacy_cleanup(self):
        # Create pre-existing AIM VRF.
        vrf = aim_resource.VRF(tenant_name='common', name='v1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, vrf)

        # Create pre-existing AIM L3Outside.
        l3out = aim_resource.L3Outside(
            tenant_name='common', name='l1', vrf_name='v1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, l3out)

        # Create pre-existing AIM ExternalNetwork.
        ext_net = aim_resource.ExternalNetwork(
            tenant_name='common', l3out_name='l1', name='n1', monitored=True)
        self.aim_mgr.create(self.aim_ctx, ext_net)

        # Create external network.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l1/instP-n1'}}
        ext_net = self._make_network(
            self.fmt, 'ext_net', True, tenant_id='tenant_1',
            arg_list=self.extension_attributes, **kwargs)['network']
        ext_net_id = ext_net['id']

        # Create router using external network.
        kwargs = {'external_gateway_info': {'network_id': ext_net_id}}
        router = self._make_router(
            self.fmt, self._tenant_id, 'router1',
            arg_list=self.extension_attributes, **kwargs)['router']
        router_id = router['id']

        # Create internal network and subnet.
        int_net_resp = self._make_network(self.fmt, 'net1', True)
        int_net = int_net_resp['network']
        int_net_id = int_net['id']
        int_subnet = self._make_subnet(
            self.fmt, int_net_resp, '10.0.1.1', '10.0.1.0/24')['subnet']
        int_subnet_id = int_subnet['id']

        # Add internal subnet to router.
        self.l3_plugin.add_router_interface(
            n_context.get_admin_context(), router_id,
            {'subnet_id': int_subnet_id})

        # Validate just to make sure everything is OK before creating
        # legacy plugin's SNAT-related resources.
        self._validate()

        # Create legacy plugin's SNAT-related Neutron network.
        leg_net_resp = self._make_network(
            self.fmt,
            'host-snat-network-for-internal-use-' + ext_net_id, False)
        leg_net = leg_net_resp['network']
        leg_net_id = leg_net['id']

        # Create legacy plugin's SNAT-related Neutron subnet.
        leg_subnet = self._make_subnet(
            self.fmt, leg_net_resp, '66.66.66.1', '66.66.66.0/24',
            enable_dhcp=False)['subnet']
        leg_subnet_id = leg_subnet['id']
        data = {'subnet': {'name': 'host-snat-pool-for-internal-use'}}
        leg_subnet = self._update('subnets', leg_subnet_id, data)['subnet']

        # Create legacy plugin's SNAT-related Neutron port.
        fixed_ips = [{'subnet_id': leg_subnet_id, 'ip_address': '66.66.66.5'}]
        leg_port = self._make_port(
            self.fmt, leg_net_id, fixed_ips=fixed_ips,
            name='host-snat-pool-port-for-internal-use',
            device_owner='host-snat-pool-port-device-owner-internal-use'
        )['port']
        leg_port_id = leg_port['id']

        # Delete all networks' mapping and extension records to
        # simulate migration use case.
        net_ids = [ext_net_id, int_net_id, leg_net_id]
        (self.db_session.query(db.NetworkMapping).
         filter(db.NetworkMapping.network_id.in_(net_ids)).
         delete(synchronize_session=False))
        (self.db_session.query(ext_db.NetworkExtensionDb).
         filter(ext_db.NetworkExtensionDb.network_id.in_(net_ids)).
         delete(synchronize_session=False))

        # Delete all subnets' extension records to simulate migration
        # use case.
        subnet_ids = [int_subnet_id, leg_subnet_id]
        (self.db_session.query(ext_db.SubnetExtensionDb).
         filter(ext_db.SubnetExtensionDb.subnet_id.in_(subnet_ids)).
         delete(synchronize_session=False))

        # Test migration.
        cfg.CONF.set_override(
            'migrate_ext_net_dns',
            {ext_net_id: 'uni/tn-common/out-l1/instP-n1'},
            group='ml2_apic_aim')
        self._validate_repair_validate()

        # Ensure legacy plugin's SNAT-related resources are gone.
        self._show(
            'ports', leg_port_id,
            expected_code=webob.exc.HTTPNotFound.code)
        self._show(
            'subnets', leg_subnet_id,
            expected_code=webob.exc.HTTPNotFound.code)
        self._show(
            'networks', leg_net_id,
            expected_code=webob.exc.HTTPNotFound.code)

        # Ensure new SNAT subnet was properly created on actual
        # external network.
        ext_subnets = self._show('networks', ext_net_id)['network']['subnets']
        self.assertEqual(1, len(ext_subnets))
        ext_subnet = self._show('subnets', ext_subnets[0])['subnet']
        self.assertEqual(leg_subnet['cidr'], ext_subnet['cidr'])
        self.assertEqual(leg_subnet['gateway_ip'], ext_subnet['gateway_ip'])
        self.assertEqual(leg_subnet['enable_dhcp'], ext_subnet['enable_dhcp'])
        self.assertEqual('SNAT host pool', ext_subnet['name'])
        self.assertTrue(ext_subnet['apic:snat_host_pool'])
        self.assertEqual(ext_net['project_id'], ext_subnet['project_id'])


class TestGbpMapping(AimValidationTestCase):

    def setUp(self):
        super(TestGbpMapping, self).setUp()

    def test_l3_policy(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create L3P.
        self.create_l3_policy()

        # Test that validation fails.
        self._validate_unrepairable()

    def test_l2_policy(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create L2P.
        l2p = self.create_l2_policy()['l2_policy']

        # Dissassociate and delete the implicitly-created L3P.
        self.db_session.query(gpdb.L2Policy).filter_by(id=l2p['id']).update(
            {'l3_policy_id': None})
        self.delete_l3_policy(l2p['l3_policy_id'])

        # Test that validation fails.
        self._validate_unrepairable()

    def test_policy_target_group(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create PTG.
        self.create_policy_target_group()

        # Dissassociating and deleting the implicitly-created L3P and
        # L2P would require removing the router interface that has
        # been created, which is not worth the effort for this
        # temporary test implementation. Manual inspection of the
        # validation output shows that validation is failing due to
        # the PTG, as well as the other resources.

        # Test that validation fails.
        self._validate_unrepairable()

    def test_policy_target(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create PTG.
        ptg = self.create_policy_target_group()['policy_target_group']

        # Create PT.
        self.create_policy_target(policy_target_group_id=ptg['id'])

        # Dissassociating and deleting the PTG, L3P and L2P is not
        # worth the effort for this temporary test
        # implementation. Manual inspection of the validation output
        # shows that validation is failing due to the PT, as well as
        # the other resources.

        # Test that validation fails.
        self._validate_unrepairable()

    def test_application_policy_group(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create APG.
        self.create_application_policy_group()

        # Test that validation fails.
        self._validate_unrepairable()

    def test_policy_classifier(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create PC.
        self.create_policy_classifier()

        # Test that validation fails.
        self._validate_unrepairable()

    def test_policy_rule_set(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create PRS.
        self.create_policy_rule_set()

        # Test that validation fails.
        self._validate_unrepairable()

    def test_external_segment(self):
        # REVISIT: Test validation of actual mapping once
        # implemented. No AIM resources are created directly, but
        # external_routes maps to the cisco_apic.EXTERNAL_CIDRS
        # network extension.

        # Create external network and subnet.
        kwargs = {'router:external': True,
                  'apic:distinguished_names':
                  {'ExternalNetwork': 'uni/tn-common/out-l1/instP-n1'}}
        net_resp = self._make_network(
            self.fmt, 'ext_net', True, arg_list=self.extension_attributes,
            **kwargs)
        subnet = self._make_subnet(
            self.fmt, net_resp, '10.0.0.1', '10.0.0.0/24')['subnet']

        # Create ES.
        self.create_external_segment(
            subnet_id=subnet['id'],
            external_routes=[{'destination': '129.0.0.0/24', 'nexthop': None}])

        # Test that validation fails.
        self._validate_unrepairable()

    def test_external_policy(self):
        # REVISIT: Test validation of actual mapping once implemented.

        # Create EP.
        self.create_external_policy()

        # Test that validation fails.
        self._validate_unrepairable()


class TestSfcMapping(test_aim_sfc_driver.TestAIMServiceFunctionChainingBase,
                     AimValidationTestMixin):

    def setUp(self):
        super(TestSfcMapping, self).setUp()
        self.av_mgr = av.ValidationManager()
        self.aim_ctx = aim_context.AimContext(self.db_session)

    def test_flow_classifier(self):
        # REVISIT: Test validation of actual mapping once
        # implemented. This resource is currently not mapped to AIM
        # until used in a port chain, but there are plans to map it
        # more proactively.

        # Create FC.
        self._create_simple_flowc()

        # Test that validation fails.
        self._validate_unrepairable()

    def test_port_port_pair_group(self):
        # REVISIT: Test validation of actual mapping once
        # implemented. This resource is currently not mapped to AIM
        # until used in a port chain, but there are plans to map it
        # more proactively.

        # Create PPG.
        self._create_simple_ppg(pairs=1)

        # Test that validation fails.
        self._validate_unrepairable()

    def test_port_chain(self):
        # REVISIT: Test validation of actual mapping once
        # implemented.

        # Create PC (along with PPG and FC).
        self._create_simple_port_chain(ppgs=1)

        # Deleting the PPG and FC, if possible, would ensure that the
        # PC itself is causing validation to fail, but is not worth
        # the effort for this temporary test implementation. Manual
        # inspection of the validation output shows that validation is
        # failing due to the PC, as well as the other resources.

        # Test that validation fails.
        self._validate_unrepairable()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nfp_db
Revision ID: 54ee8e8d205a
Revises: 98862f2d814e
"""


# revision identifiers, used by Alembic.
revision = '54ee8e8d205a'
down_revision = '98862f2d814e'


from alembic import op
import sqlalchemy as sa

from gbpservice.nfp.common import constants as nfp_constants


def upgrade():

    op.create_table(
        'nfp_port_infos',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('port_model',
                  sa.Enum(nfp_constants.NEUTRON_PORT,
                          nfp_constants.GBP_PORT,
                          name='port_model'),
                  nullable=False),
        sa.Column('port_classification',
                  sa.Enum(nfp_constants.PROVIDER,
                          nfp_constants.CONSUMER,
                          nfp_constants.MANAGEMENT,
                          nfp_constants.MONITOR,
                          name='port_classification'),
                  nullable=False),
        sa.Column('port_role',
                  sa.Enum(nfp_constants.ACTIVE_PORT,
                          nfp_constants.STANDBY_PORT,
                          nfp_constants.MASTER_PORT,
                          name='port_role'),
                  nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'nfp_network_infos',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('network_model',
                  sa.Enum(nfp_constants.NEUTRON_NETWORK,
                          nfp_constants.GBP_NETWORK,
                          name='network_model'),
                  nullable=False),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'nfp_network_functions',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=1024), nullable=True),
        sa.Column('status', sa.String(length=50), nullable=True),
        sa.Column('status_description', sa.String(length=4096), nullable=True),
        sa.Column('service_id', sa.String(length=36), nullable=False),
        sa.Column('service_chain_id', sa.String(length=36), nullable=False),
        sa.Column('service_profile_id', sa.String(length=36), nullable=True),
        sa.Column('service_config', sa.TEXT(), nullable=True),
        sa.Column('config_policy_id', sa.String(length=36), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'nfp_network_function_devices',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=50), nullable=True),
        sa.Column('status_description', sa.String(length=4096), nullable=True),
        sa.Column('mgmt_ip_address', sa.String(length=36), nullable=True),
        sa.Column('mgmt_port_id',
                  sa.String(length=36),
                  nullable=True),
        sa.Column('monitoring_port_id',
                  sa.String(length=36),
                  nullable=True),
        sa.Column('monitoring_port_network',
                  sa.String(length=36),
                  nullable=True),
        sa.Column('service_vendor', sa.String(length=36), nullable=True),
        sa.Column('max_interfaces', sa.Integer(), nullable=True),
        sa.Column('reference_count', sa.Integer(), nullable=True),
        sa.Column('interfaces_in_use', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['mgmt_port_id'],
                                ['nfp_port_infos.id'],
                                ondelete='SET NULL',
                                name='nfp_nfd_mgmt_fk_port_info'),
        sa.ForeignKeyConstraint(['monitoring_port_network'],
                                ['nfp_network_infos.id'],
                                ondelete='SET NULL',
                                name='nfp_nfd_fk_net_info'),
        sa.ForeignKeyConstraint(['monitoring_port_id'],
                                ['nfp_port_infos.id'],
                                ondelete='SET NULL',
                                name='nfp_nfd_mon_fk_port_info'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'nfp_network_function_instances',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=50), nullable=True),
        sa.Column('status_description', sa.String(length=4096), nullable=True),
        sa.Column('ha_state', sa.String(length=50), nullable=True),
        sa.Column('network_function_id', sa.String(length=36), nullable=True),
        sa.Column('network_function_device_id',
                  sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['network_function_device_id'],
                                ['nfp_network_function_devices.id'],
                                ondelete='SET NULL',
                                name='nfp_nfi_fk_nfd'),
        sa.ForeignKeyConstraint(['network_function_id'],
                                ['nfp_network_functions.id'],
                                ondelete='SET NULL',
                                name='nfp_nfi_fk_nf'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'nfp_nfi_dataport_associations',
        sa.Column('network_function_instance_id',
                  sa.String(length=36),
                  nullable=True),
        sa.Column('data_port_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['network_function_instance_id'],
                                ['nfp_network_function_instances.id'],
                                name='nfp_nfi_dp_assoc_fk_nfi'),
        sa.ForeignKeyConstraint(['data_port_id'], ['nfp_port_infos.id'],
                                ondelete='CASCADE',
                                name='nfp_nfi_dp_assoc_fk_port_info'),
        sa.PrimaryKeyConstraint('network_function_instance_id', 'data_port_id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg as oslo_config

from oslo_service import loopingcall as oslo_looping_call
from oslo_service import periodic_task as oslo_periodic_task
from oslo_service import service as oslo_service

from neutron.agent import rpc as n_agent_rpc
from neutron.common import rpc as n_rpc

from neutron import context as n_context

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)

n_rpc.init(oslo_config.CONF)

"""RPC Service class.

    This class is implemented from neutron.common.rpc
    because neutron limit itself to single rpc manager
    per topic while oslo allows multiple.
    Using this class, one can register multiple rpc
    managers with oslo.
"""


class RpcService(oslo_service.Service):

    def __init__(self, topic, managers=None):
        # Not taking host and serializer as neutron rpc does not support
        # passing any host and serializer- it picks up the host from
        # 'cfg.CONF.host' and consider serializer as None

        super(RpcService, self).__init__()
        self.topic = topic
        self.endpoints = managers or []

    def start(self):
        LOG.debug("RPCAgent listening on topic=%s", self.topic)
        super(RpcService, self).start()

        self.conn = n_rpc.create_connection()
        LOG.debug("Creating Consumer connection for Service %s",
                  self.topic)

        self.conn.create_consumer(self.topic, self.endpoints)

        # Hook to allow the manager to do other initializations after
        # the rpc connection is created.
        for manager in self.endpoints:
            if callable(getattr(manager, 'initialize_service_hook', None)):
                self.manager.initialize_service_hook(self)

        # Consume from all consumers in threads
        self.conn.consume_in_threads()

    def stop(self):
        # Try to shut the connection down, but if we get any sort of
        # errors, go ahead and ignore them.. as we're shutting down anyway
        try:
            self.conn.close()
        except Exception:
            pass
        super(RpcService, self).stop()


"""Wrapper class for Neutron RpcAgent definition.

    NFP modules will use this class for the agent definition.
    Associates the state reporting of agent to ease
    the usage for modules.
"""


class RpcAgent(object):

    def __init__(
            self, sc, host=None,
            topic=None, manager=None, report_state=None, priority=0):
        # report_state =
        #   {<agent_state_keys>, 'plugin_topic': '', 'report_interval': ''}

        self.sc = sc
        # The argument 'host' will be ignored as neutron rpc does not
        # support passing any host -it picks up from 'cfg.CONF.host'
        self.host = oslo_config.CONF.host
        self.topic = topic
        self.manager = manager
        self.priority = priority
        # Check if the agent needs to report state
        if report_state:
            self._report_state = ReportState(report_state)

    def report_state(self):
        if hasattr(self, '_report_state'):
            LOG.debug("Agent (%s) reporting state",
                      (self.identify()))
            self._report_state.report()

    def identify(self):
        return "(host=%s,topic=%s)" % (self.host, self.topic)


"""This class implements the state reporting for neutron *aaS agents

    One common place of handling of reporting logic.
    Each nfp module just need to register the reporting data and
    plugin topic.
"""


class ReportState(object):

    def __init__(self, data):
        self._n_context = n_context.get_admin_context_without_session()
        self._data = data
        self._topic = data.pop('plugin_topic', None)
        self._interval = data.pop('report_interval', 0)
        self._state_rpc = n_agent_rpc.PluginReportStateAPI(
            self._topic)

    def report(self):
        try:
            LOG.debug("Reporting state with data (%s)",
                      (self._data))
            self._state_rpc.report_state(self._n_context, self._data)
            self._data.pop('start_flag', None)
        except AttributeError:
            # This means the server does not support report_state
            message = "Neutron server does not support state report."
            "Agent State reporting will be disabled"
            LOG.info(message)
            return
        except Exception:
            message = "Stopped reporting agent state!"
            LOG.exception(message)

"""Periodic task to report neutron *aaS agent state.

    Derived from oslo periodic task, to report the agents state
    if any, to neutron *aaS plugin.
"""


class ReportStateTask(oslo_periodic_task.PeriodicTasks):

    def __init__(self, conf, controller):
        super(ReportStateTask, self).__init__(conf)
        self._controller = controller
        # Start a looping at the defined pulse
        pulse = oslo_looping_call.FixedIntervalLoopingCall(
            self.run_periodic_tasks, None, None)
        pulse.start(
            interval=5, initial_delay=None)

    @oslo_periodic_task.periodic_task(spacing=10)
    def report_state(self, context):
        # trigger the state reporting
        self._controller.report_state()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from neutron.common import config as neutron_config  # noqa

from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.tests.unit import common as cm
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    extensions as test_ext)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_grouppolicy_plugin as test_plugin)


class ExtensionDriverTestBase(test_plugin.GroupPolicyPluginTestCase):
    _extension_drivers = ['test']
    _extension_path = os.path.dirname(os.path.abspath(test_ext.__file__))

    def setUp(self, policy_drivers=None, core_plugin=None,
              l3_plugin=None, ml2_options=None,
              sc_plugin=None, trunk_plugin=None):
        config.cfg.CONF.set_override('extension_drivers',
                                     self._extension_drivers,
                                     group='group_policy')
        if self._extension_path:
            config.cfg.CONF.set_override(
                'api_extensions_path', self._extension_path)
        super(ExtensionDriverTestBase, self).setUp(
            core_plugin=core_plugin, l3_plugin=l3_plugin,
            ml2_options=ml2_options, sc_plugin=sc_plugin,
            trunk_plugin=trunk_plugin)


class ExtensionDriverTestCase(ExtensionDriverTestBase):

    def test_pt_attr(self):
        # Test create with default value.
        pt = self.create_policy_target()
        policy_target_id = pt['policy_target']['id']
        val = pt['policy_target']['pt_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_targets', policy_target_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_targets')
        val = res['policy_targets'][0]['pt_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        pt = self.create_policy_target(pt_extension="abc")
        policy_target_id = pt['policy_target']['id']
        val = pt['policy_target']['pt_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_targets', policy_target_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_target': {'pt_extension': "def"}}
        req = self.new_update_request('policy_targets', data, policy_target_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_targets', policy_target_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target']['pt_extension']
        self.assertEqual("def", val)

    def test_ptg_attr(self):
        # Test create with default value.
        ptg = self.create_policy_target_group()
        policy_target_group_id = ptg['policy_target_group']['id']
        val = ptg['policy_target_group']['ptg_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_target_groups',
                                    policy_target_group_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_target_groups')
        val = res['policy_target_groups'][0]['ptg_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        ptg = self.create_policy_target_group(ptg_extension="abc")
        policy_target_group_id = ptg['policy_target_group']['id']
        val = ptg['policy_target_group']['ptg_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_target_groups',
                                    policy_target_group_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_target_group': {'ptg_extension': "def"}}
        req = self.new_update_request('policy_target_groups', data,
                                      policy_target_group_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_target_groups',
                                    policy_target_group_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_target_group']['ptg_extension']
        self.assertEqual("def", val)

    def test_l2p_attr(self):
        # Test create with default value.
        l2p = self.create_l2_policy()
        l2_policy_id = l2p['l2_policy']['id']
        val = l2p['l2_policy']['l2p_extension']
        self.assertIsNone(val)
        req = self.new_show_request('l2_policies', l2_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('l2_policies')
        val = res['l2_policies'][0]['l2p_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        l2p = self.create_l2_policy(l2p_extension="abc")
        l2_policy_id = l2p['l2_policy']['id']
        val = l2p['l2_policy']['l2p_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('l2_policies', l2_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'l2_policy': {'l2p_extension': "def"}}
        req = self.new_update_request('l2_policies', data, l2_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('l2_policies', l2_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l2_policy']['l2p_extension']
        self.assertEqual("def", val)

    def test_l3p_attr(self):
        # Test create with default value.
        l3p = self.create_l3_policy()
        l3_policy_id = l3p['l3_policy']['id']
        val = l3p['l3_policy']['l3p_extension']
        self.assertIsNone(val)
        req = self.new_show_request('l3_policies', l3_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('l3_policies')
        val = res['l3_policies'][0]['l3p_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        l3p = self.create_l3_policy(l3p_extension="abc")
        l3_policy_id = l3p['l3_policy']['id']
        val = l3p['l3_policy']['l3p_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('l3_policies', l3_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'l3_policy': {'l3p_extension': "def"}}
        req = self.new_update_request('l3_policies', data, l3_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('l3_policies', l3_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['l3_policy']['l3p_extension']
        self.assertEqual("def", val)

    def test_pc_attr(self):
        # Test create with default value.
        pc = self.create_policy_classifier()
        policy_classifier_id = pc['policy_classifier']['id']
        val = pc['policy_classifier']['pc_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_classifiers', policy_classifier_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_classifiers')
        val = res['policy_classifiers'][0]['pc_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        pc = self.create_policy_classifier(pc_extension="abc")
        policy_classifier_id = pc['policy_classifier']['id']
        val = pc['policy_classifier']['pc_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_classifiers', policy_classifier_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_classifier': {'pc_extension': "def"}}
        req = self.new_update_request('policy_classifiers', data,
                                      policy_classifier_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_classifiers', policy_classifier_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_classifier']['pc_extension']
        self.assertEqual("def", val)

    def test_pa_attr(self):
        # Test create with default value.
        pa = self.create_policy_action()
        policy_action_id = pa['policy_action']['id']
        val = pa['policy_action']['pa_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_actions', policy_action_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_actions')
        val = res['policy_actions'][0]['pa_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        pa = self.create_policy_action(pa_extension="abc")
        policy_action_id = pa['policy_action']['id']
        val = pa['policy_action']['pa_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_actions', policy_action_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_action': {'pa_extension': "def"}}
        req = self.new_update_request('policy_actions', data, policy_action_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_actions', policy_action_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_action']['pa_extension']
        self.assertEqual("def", val)

    def test_pr_attr(self):
        # Create necessary parameters.
        classifier = self.create_policy_classifier(
            name="class1", protocol="tcp", direction="out",
            port_range="50:100")
        classifier_id = classifier['policy_classifier']['id']

        # Test create with default value.
        pr = self.create_policy_rule(policy_classifier_id=classifier_id)
        policy_rule_id = pr['policy_rule']['id']
        val = pr['policy_rule']['pr_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_rules', policy_rule_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_rules')
        val = res['policy_rules'][0]['pr_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        pr = self.create_policy_rule(policy_classifier_id=classifier_id,
                                     pr_extension="abc")
        policy_rule_id = pr['policy_rule']['id']
        val = pr['policy_rule']['pr_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_rules', policy_rule_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_rule': {'pr_extension': "def"}}
        req = self.new_update_request('policy_rules', data, policy_rule_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_rules', policy_rule_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule']['pr_extension']
        self.assertEqual("def", val)

    def test_prs_attr(self):
        # Test create with default value.
        prs = self.create_policy_rule_set(policy_rules=[])
        policy_rule_set_id = prs['policy_rule_set']['id']
        val = prs['policy_rule_set']['prs_extension']
        self.assertIsNone(val)
        req = self.new_show_request('policy_rule_sets', policy_rule_set_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('policy_rule_sets')
        val = res['policy_rule_sets'][0]['prs_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        prs = self.create_policy_rule_set(policy_rules=[], prs_extension="abc")
        policy_rule_set_id = prs['policy_rule_set']['id']
        val = prs['policy_rule_set']['prs_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('policy_rule_sets', policy_rule_set_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'policy_rule_set': {'prs_extension': "def"}}
        req = self.new_update_request('policy_rule_sets', data,
                                      policy_rule_set_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('policy_rule_sets', policy_rule_set_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['policy_rule_set']['prs_extension']
        self.assertEqual("def", val)

    def test_nsp_attr(self):
        # Test create with default value.
        nsp = self.create_network_service_policy()
        network_service_policy_id = nsp['network_service_policy']['id']
        val = nsp['network_service_policy']['nsp_extension']
        self.assertIsNone(val)
        req = self.new_show_request('network_service_policies',
                                    network_service_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list('network_service_policies')
        val = res['network_service_policies'][0]['nsp_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        nsp = self.create_network_service_policy(nsp_extension="abc")
        network_service_policy_id = nsp['network_service_policy']['id']
        val = nsp['network_service_policy']['nsp_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request('network_service_policies',
                                    network_service_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {'network_service_policy': {'nsp_extension': "def"}}
        req = self.new_update_request('network_service_policies', data,
                                      network_service_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("def", val)
        req = self.new_show_request('network_service_policies',
                                    network_service_policy_id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res['network_service_policy']['nsp_extension']
        self.assertEqual("def", val)

    def test_es_attr(self):
        self._test_attr('external_segment')

    def test_ep_attr(self):
        self._test_attr('external_policy')

    def test_np_attr(self):
        self._test_attr('nat_pool')

    def _test_attr(self, type):
        # Test create with default value.
        acronim = _acronim(type)
        plural = cm.get_resource_plural(type)
        obj = getattr(self, 'create_%s' % type)()
        id = obj[type]['id']
        val = obj[type][acronim + '_extension']
        self.assertIsNone(val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertIsNone(val)

        # Test list.
        res = self._list(plural)
        val = res[plural][0][acronim + '_extension']
        self.assertIsNone(val)

        # Test create with explict value.
        kwargs = {acronim + '_extension': "abc"}
        obj = getattr(self, 'create_%s' % type)(**kwargs)
        id = obj[type]['id']
        val = obj[type][acronim + '_extension']
        self.assertEqual("abc", val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("abc", val)

        # Test update.
        data = {type: {acronim + '_extension': "def"}}
        req = self.new_update_request(plural, data, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("def", val)
        req = self.new_show_request(plural, id)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        val = res[type][acronim + '_extension']
        self.assertEqual("def", val)


def _acronim(type):
    return ''.join([x[0] for x in type.split('_')])
